#!/usr/bin/perl                    #-w for warnings

package meta_pred;
use Bio::SeqIO;
use Bio::Seq;
use Bio::Tools::CodonTable;
use YACOP::gene_pred::CodonTable;
use YACOP::gene_pred::DataHandle;
use YACOP::gene_pred::Element;
use YACOP::gene_pred::ResultSet;
use YACOP::gene_pred::Result;
use YACOP::gene_pred::GenebankParser;
use YACOP::gene_pred::GlimmerParser;     
use YACOP::gene_pred::CriticaParser;
use YACOP::gene_pred::OrpheusParser;
use YACOP::gene_pred::ZCurveParser;
use YACOP::gene_pred::GlimmerRBSParser;
use YACOP::gene_pred::QuickResult;

#=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
# author mt
# predicted ORFs are written in ascending order to STDOUT additionally two
# files containing a selection of predicted genes resp. to the mode given 
# with param -m are created (PREFIX.tbl, PREFIX.fasta).  
# output: STDOUT (PREFIX.sum_out) contains all results (indep. of mode)
#         PREFIX.tbl contains the contigs (id\tnama_start_stop)
#         PREFIX.fasta contains the extracted amino acid sequences in 
#                      multiple fasta-format.
#         STDERR (meta_pred.msg) contains in concat mode all removed 
#                predicted coding regions
#
# in concat mode an indexfile, should be given to meta_pred with param -i
#=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
$|=1; #Output is not buffered
my $USAGE = "\a\n++++++++++++++++++++++++++++++++++\nusage: meta_pred.pl ".
    "-f fasta [-g glimmer -o orpheus \n"
    ."-c critica  -gr glimmerRBS -gb genebank] -i indexfile\n".
    " -d outputdir -p prefix -m outputmode -l minlength] \n>> output\n".
            "++++++++++++++++++++++++++++++++++\n";
#--flags-----------------------------------------------------
my $F_GLIMMER = 'g';
my $F_GLIMMER_RBS = 'gr'; 
my $F_CRITICA = 'c';
my $F_ORPHEUS = 'o';
my $F_ZCURVE = 'z';
my $F_GENEBANK = 'gb';
my $F_GENOM = 'f';
my $F_PREFIX = 'p';
my $F_MODE = 'm';
my $F_MINLENGTH= 'l';
my $F_OUTPUT_DIR = 'd';
my $F_INDEX= 'i';
#--------modes------------------------------------------------------
my $M_CRIT_ORP_GL = 'crit_orp_gl';        #critica + (orpheus&&glimmer) 
                                          #start: c or o
my $M_CRIT_ZC_GL_ORP = 'crit_zc_gl_orp';  #crit +(orp&&gl&&zc)
                                          #start: c or zc
my $M_ZC_CRIT_ORP_GL = 'zc_crit_orp_gl';  #crit +(orp&&gl&&zc)
                                          #start: zc or c
my $M_ZC_CRIT_GL = 'zc_crit_gl';          #crit +(gl&&zc) start: zc or c
my $M_CRIT_ZC_GL = 'crit_zc_gl';          #crit +(gl&&zc) start: c or zc
my $M_ZC_GL= 'zc_gl';                     #zc&&gl start: zc
my $M_GL = 'gl';                          #only glimmer output is considered
#-------------------------------------------------------------------

my $STOP='stop';
my $COML='compl';
my $GBK_ANNO='gbk_anno';

my $OUTPUT_DIR = $ENV{HOME};           
my (%HANDLE,  $GENOM, @INDEX, @val_out, $PREFIX, $MODE);
my $MINLENGTH=0;

#-----------------------------MAIN-PROGRAM---------------------------------
if (@ARGV<2){print $USAGE; exit(1)}
$INDEX[0]=-1; #-1 if no index file is given during invokation
&eval_args();
$r = &setup_result();
&trim_result($r) if($INDEX[0]!=-1);      
&output_all($r);
my $quick = &quick_results($r);
&write_table($quick);
&write_multiple_fasta($quick);
exit(0);

#-----------------------------SUBROUTINES----------------------------------
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# generates the resultset, containing all predicted ORFs.
# @return  ResultSet
#------------------------
sub setup_result{
    my @handles = keys %HANDLE;
    my $result_set = ResultSet->new;
    foreach(@handles){    	
	for(my $i=1; $i<=$HANDLE{$_}->counter; $i++){
	    my $e = $HANDLE{$_}->get_element($i);
	    if ($e){
		$result_set->add_result($_, $e->end, $i);}
	}
    }
    return $result_set;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# extracting predicted coding regions at contig transitions (in concat mode)
# all predicted coding regions, which were removed from resultset are 
# listed to STDERR (if started with combinate STDERR is piped to 
# ~/data/longtime/meta_pred.msg
# @param ResultSet
#------------------------
sub trim_result{
    my $set = shift;
    if ($INDEX[0]!=-1){
	
	print STDERR "--------------------------------------------------\n"
	    ."removed predicted coding regions:\n"
	    ."compl\tstop\tstart\tparser\n\n"
	     ."--------------------------------------------------\n";
	
	@INDEX = sort {$a <=> $b} @INDEX;
	@handle_type = keys %HANDLE;
	@result_ids=$set->get_all_results;
	foreach(@result_ids){                   #id=stop
	    my $stop = $_; 
	    my $res = $set->get_result($_);
	    foreach(@handle_type){
		$h=$_;
		my $tmp_id= $res->value($h);
		$tmp = $HANDLE{$h}->get_element($tmp_id);
		if ($tmp){
		    if ($tmp->start < $stop){
			foreach(@INDEX){
			    if($_ >= $tmp->start && $_ <= $stop){
				print STDERR "F\t".$stop."\t".$tmp->start.
				    "\t".$h."\n";
				$set->remove_element($stop, $h);
				$HANDLE{$h}->remove_element($tmp_id);
				last;
			    }
			    last if($_> $stop);
			}
		    }
		    else{
			foreach(@INDEX){
			    if($_ <= $tmp->start && $_ >= $stop){
				print STDERR "R\t".$stop."\t"
				    .$tmp->start."\t".$h."\n";
				$set->remove_element($stop, $h);
				$HANDLE{$h}->remove_element($tmp_id);
				last;
			    }
			    last if($_>$tmp->start);}
		    }
		}
	    } 
	} 
	print STDERR "--------------------------------------------------\n";
    }
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# generates a light resultset containing only id, start, stop of ORFs 
# >$MINLENGTH
# @param ResultSet
# @return array reference holding QuickResults
#------------------------
sub quick_results{
    my $set = shift;
    my ($id, $start, $end, %starts, $tmp, @quick, @rest);
    my $counter=0;
    my $C_OL=0;
    @handle_type = keys %HANDLE;
    @result_ids=$set->get_all_results;
    @result_ids = sort {$a <=> $b} @result_ids;
    foreach(@result_ids){      #result id is stop position
	my $id = $_;
	$res = $set->get_result($id);
	undef $start;
	undef $end;
	undef %starts;
	if ( $res && $res != -1){
	    $C_OL=0;
	    foreach(@handle_type){
		$tmp = $res->value($_);
		if ($tmp) {
		    $tmp = $HANDLE{$_}->get_element($tmp);
		    if ($_ eq CriticaParser){
			$C_OL=1;}
		    if ($tmp->complement=~/R/i){
			$end =$id+1; }
		    else{$end=$id-1;}
		    $starts{$_}=$tmp->start;
		    undef $tmp; 
		}
	    }
	    $start = &eval_start_mode(%starts);
	    if ( $start && $start != 0 
		&& ($C_OL==1||&length_ok($start, $end))){
		$quick[$counter++] = QuickResult->new($counter, $start, $end); 
	    }
	    else{print STDERR "undefined mode\n"}
	}
    } #end for ids
    return \@quick;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#@param start
#@param end
#@return 1 if length is ok, o if not.
#---------------------------
sub length_ok(){
    my ($start, $end)=@_;
    if ($start>$end){
        $l=$start-$end;}
    else{$l=$end-$start;}
    if ($l>=$MINLENGTH){
        return 1;}
    else {return 0;}
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# writing results (id\tcontig_start_end) to outputfile.
# id according to regular expression /^R([A-Z]{2,3})\d{5,6}/
# @param QuickResult array
#------------------------
sub write_table{
    my $results = shift;
    my $file = $OUTPUT_DIR."/".$PREFIX.".tbl";
    open(e_table, ">".$file) 
	|| die "cannot write table file ".$file."\n";
    for(my $rid=0; $rid < @{$results}; $rid++){ 
	printf e_table "R%s%06d\t%s_%d_%d\n", $PREFIX, $rid+1, 
	$GENOM->display_id, $results->[$rid]->start, 
	$results->[$rid]->stop;     
    }
    close(e_table);
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# writing results (amino acid sequence).  
# @param QuickResult array
#------------------------
sub write_multiple_fasta{
    my $results = shift;
    my $table = Bio::Tools::CodonTable->new();
    my $file = $OUTPUT_DIR."/".$PREFIX.".fasta";
    open(multi_fasta, ">".$file) 
	|| die "cannot write multiplefasta file ".$file."\n";
    for (my $rid=0; $rid < @{$results}; $rid++){
	printf multi_fasta ">R%s%06d\n", $PREFIX, $rid+1;  
	
	if($results->[$rid]->start < $results->[$rid]->stop){	    
	    for(my $i=$results->[$rid]->start;$i<$results->[$rid]->stop;
		$i+=180){
		if($i+179<$results->[$rid]->stop){
		    print multi_fasta 
			$table->translate($GENOM->subseq($i,$i+179));
		}
		else{
		    print multi_fasta $table->translate($GENOM->subseq($i, 
						     $results->[$rid]->stop));
		}
		print multi_fasta "\n";
	    }
	}
	else{
	    @nuc_seq =  split "", $GENOM->subseq($results->[$rid]->stop, 
						 $results->[$rid]->start);
	    my $rest = @nuc_seq % 60;
	    my $rev_seq;
	    for (my$y=@nuc_seq-1; $y>=0; $y--){
		$rev_seq.=CodonTable->complement_nuc($nuc_seq[$y]);
	    }
	    for (my $i=0; $i<@nuc_seq; $i+=180){
		if ($i+180<@nuc_seq){
		    print multi_fasta $table->translate(substr $rev_seq,
							$i,180);
		}
		else{
		    print multi_fasta $table->translate(substr $rev_seq, $i);
		}
		print multi_fasta "\n";
	    }
	}
    }
    close(multi_fasta);
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# selects start coordinate according to current mode and returns it.
# @param hash with predicted startpositions, 
# @return start coordinate according to the mode.
#-------------
sub eval_start_mode{
    my %starts =@_; 
    my $start=0;
    if ($MODE eq $M_CRIT_ORP_GL){                  
	if( $starts{CriticaParser}){      
	    $start = $starts{CriticaParser};
	}
	elsif( $starts{OrpheusParser} 
              &&  $starts{GlimmerRBSParser}){
	    $start= $starts{OrpheusParser};
	}
    }    
    elsif($MODE eq $M_ZC_CRIT_ORP_GL){
	if ( $starts{CriticaParser}){
	    if ( $starts{ZCurveParser}){
		$start = $starts{ZCurveParser};
	    }
	    else{
		$start=$starts{CriticaParser};
	    }
	}
	elsif( $starts{ZCurveParser}
	      &&  $starts{OrpheusParser}
	      &&  $starts{GlimmerRBSParser}){
	    $start=$starts{ZCurveParser};
	}
    }
    elsif ($MODE eq $M_CRIT_ORP_GL_ZC){    
	if( $starts{CriticaParser}){     
	    $start = $starts{CriticaParser};
	}
	elsif( $starts{OrpheusParser}
	      &&  $starts{ZCurveParser}
	      &&  $starts{GlimmerRBSParser}){
	    $start= $starts{OrpheusParser};
	}
    }
    elsif ($MODE eq $M_CRIT_ZC_GL_ORP){    
	if( $starts{CriticaParser}){     
	    $start = $starts{CriticaParser};
	}
	elsif( $starts{OrpheusParser}
	      &&  $starts{ZCurveParser}
	      &&  $starts{GlimmerRBSParser}){
	    $start= $starts{ZCurveParser};
	}
    }
    elsif ($MODE eq $M_CRIT_ORP){
	if( $starts{CriticaParser}){
	    $start = $starts{CriticaParser};
	}
	elsif( $starts{OrpheusParser}){
	    $start = $starts{OrpheusParser};
	}
    }
    elsif ($MODE eq $M_ZC_CRIT_GL){
	if( $starts{CriticaParser}){
	    if ( $starts{ZCurveParser}){
		$start=$starts{ZCurveParser};}
	    else{
		$start= $starts{CriticaParser};
	    }
	}
	elsif( $starts{ZCurveParser}
	      &&  $starts{GlimmerRBSParser}){
	    $start=$starts{ZCurveParser};
	}
    }
    elsif ($MODE eq $M_CRIT_ZC_GL){
	if($starts{CriticaParser}){
	    $start= $starts{CriticaParser};
	}
	elsif($starts{ZCurveParser}
	      && $starts{GlimmerRBSParser}){
	    $start=$starts{ZCurveParser}
	}
    }
    elsif ($MODE eq $M_ZC_GL){
	if($starts{ZCurveParser}
	      && $starts{GlimmerRBSParser}){
	    $start=$starts{ZCurveParser}
	}
    }
    elsif($MODE eq $M_GL){
	$start=$starts{GlimmerRBSParser} 
	if($starts{GlimmerRBSParser});
    }
    return $start;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# writing results of all tools to stdout (inclusive predicted ORFs 
# <$MINLENGTH )
#------------------------
sub output_all{
    my $set = shift;
    my ($i, @handle_type, @result_ids, $stop);
    my $bool=0;
    @handle_type = keys %HANDLE;
    @result_ids = $set->get_all_results;
    @result_ids = sort {$a <=> $b} @result_ids;
    foreach(@result_ids){
	$bool = 0;
	my %val,
	my $tstart=0;
	$res = $set->get_result($_);
	if($res && $res != -1){
	    foreach(@handle_type){
		my $tmp = $res->value($_);
		if ($tmp){
		    $tmp = $HANDLE{$_}->get_element($tmp);
#		    $val{$COMPL}=$tmp->complement;
		    $tstart=$tmp->start;
		    $bool++;
		    $val{$_."S"} = $tmp->score;
		    $val{$STOP}=$tmp->end;
		    $val{$_}=$tmp->start;
		    if ($_ eq GenebankParser->type){
			$val{$GBK_ANNO} = $tmp->comment;
			$val{$RBS} = $tmp->rbs;
                    }
		}
		else {
		    $val{$_."S"}= '-';
		    $val{$_} = '-';  }        
	    }
	    if($tstart<$val{$STOP}){$val{$COMPL}='F'}
	    else{$val{$COMPL}='R'}
	    &write_all(\%val);
	}
    }
    print STDERR "---------------------------------------------------\n";
    print STDERR "Minlength: ".$MINLENGTH." BP\n";
    foreach(@handle_type){print STDERR $_.": ".$HANDLE{$_}->counter."\n"}
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# writes given predictions to STDOUT, total number of predictions
# is written to STDERR.
#---------------------
sub write_all{
    my $val = shift;
    
    return -1 if (!$val->{$STOP});
    if ($MODE eq $M_CRIT_ORP_GL){                  
	printf "%7s %s  %7s (%3s)  %7s %9s  %7s  %s %s\n", 
	$val->{$STOP}, $val->{$COMPL}, $val->{GlimmerRBSParser->type}, 
	$val->{(GlimmerRBSParser->type)."S"}, $val->{CriticaParser->type}, 
	$val->{(CriticaParser->type)."S"}, $val->{OrpheusParser->type}, 
	$val->{GenebankParser->type}, $val->{$GBK_ANNO}
    }    
    elsif($MODE eq $M_ZC_CRIT_ORP_GL 
	  ||$MODE eq $M_CRIT_ORP_GL_ZC
	  ||$MODE eq $M_CRIT_ZC_GL_ORP){
	printf "%7s %s  %7s (%3s)  %7s %9s  %7s  %7s %7s  %s %s\n", 
	$val->{$STOP}, $val->{$COMPL}, $val->{GlimmerRBSParser->type}, 
	$val->{(GlimmerRBSParser->type)."S"}, $val->{CriticaParser->type}, 
	$val->{(CriticaParser->type)."S"}, $val->{OrpheusParser->type}, 
	$val->{ZCurveParser->type}, $val->{(ZCurveParser->type)."S"}, 
	$val->{GenebankParser->type}, $val->{$GBK_ANNO}
    }
    elsif ($MODE eq $M_CRIT_ORP){
	printf "%7s %s  %7s %9s  %7s  %s %s\n",
	$val->{$STOP}, $val->{$COMPL},$val->{CriticaParser->type}, 
	$val->{(CriticaParser->type)."S"}, $val->{OrpheusParser->type}, 
	$val->{GenebankParser->type}, $val->{$GBK_ANNO}
    }
    elsif ($MODE eq $M_ZC_CRIT_GL
	   ||$MODE eq $M_CRIT_ZC_GL 
	   || ($MODE eq $M_ZC_GL && $HANDLE{CriticaParser->type})){
	    printf "%7s %s  %7s (%3s)  %7s %9s  %7s %7s  %s %s\n", 
	    $val->{$STOP}, $val->{$COMPL}, $val->{GlimmerRBSParser->type}, 
	    $val->{(GlimmerRBSParser->type)."S"}, $val->{CriticaParser->type}, 
	    $val->{(CriticaParser->type)."S"},
	    $val->{ZCurveParser->type}, $val->{(ZCurveParser->type)."S"}, 
	    $val->{GenebankParser->type}, $val->{$GBK_ANNO}
    }
    elsif ($MODE eq $M_ZC_GL){
	printf "%7s %s  %7s (%3s)  %7s %7s  %s %s\n", 
	$val->{$STOP}, $val->{$COMPL}, $val->{GlimmerRBSParser->type}, 
	$val->{(GlimmerRBSParser->type)."S"}, $val->{ZCurveParser->type}, 
	$val->{(ZCurveParser->type)."S"}, $val->{GenebankParser->type}, 
	$val->{$GBK_ANNO}
    }
    elsif ($MODE eq $M_GL){
	if ($HANDLE{CriticaParser->type}){
	    printf "%7s %s  %7s (%3s)  %7s %9s  %s %s\n", 
	    $val->{$STOP}, $val->{$COMPL}, $val->{GlimmerRBSParser->type}, 
	    $val->{(GlimmerRBSParser->type)."S"},
	    $val->{CriticaParser->type}, $val->{(CriticaParser->type)."S"},
	    $val->{GenebankParser->type}, $val->{$GBK_ANNO}}
	else{	    
	    printf "%7s %s  %7s (%3s)  %s %s\n", 
	    $val->{$STOP}, $val->{$COMPL}, $val->{GlimmerRBSParser->type}, 
	    $val->{(GlimmerRBSParser->type)."S"}, 
	    $val->{GenebankParser->type}, $val->{$GBK_ANNO}}
    }
    else {print STDERR "undefined mode\n"}
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Fasta-input is parsed to Bio::Seq - object.
#@param filename of fasta-file (NO multiple fasta, only the first one would
#                               be taken 
#@return  Bio::Seq - object
#------------------------
sub eval_fasta{
    my $file = shift;
    my $input  = Bio::SeqIO->new('-file'=> $file, '-format'=>"FASTa");
    my $seq = $input->next_seq();
    return $seq; 
} 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Evaluates the commandline parameters and invokes parse routine of resp. 
# Parser (returns a DataHandle holding the Genes in ascending order).     
#------------------------
sub eval_args{
    my @tmp;
    while (@tmp = splice @ARGV, 0, 2){	
	if ($tmp[0] =~ s/-//){
	    if ($tmp[0] eq $F_GENOM){
		$GENOM = &eval_fasta($tmp[1]);}
	    elsif ($tmp[0] eq $F_PREFIX){$PREFIX = $tmp[1];}
	    elsif ($tmp[0] eq $F_MODE){$MODE = $tmp[1];}
	    elsif ($tmp[0] eq $F_OUTPUT_DIR){$OUTPUT_DIR = $tmp[1];}
	    elsif ($tmp[0] eq $F_GLIMMER){
		$HANDLE{GlimmerParser->type} = GlimmerParser->parse($tmp[1]);
		$HANDLE{GlimmerParser->type}->sort_elements;
	    }
	    elsif ($tmp[0] eq $F_ORPHEUS){
		$HANDLE{OrpheusParser->type} = OrpheusParser->parse($tmp[1]);
		$HANDLE{OrpheusParser->type}->sort_elements;
	    }
	    elsif ($tmp[0] eq $F_INDEX){
		open ix, $tmp[1] || die "cannot open index-file: "
		    .$tmp[1]."\n";
		undef $INDEX[0];
		while (<ix>){
		    my @t=split;
		    @t=split '_', $t[@t-1];
		    push @INDEX, $t[1];
		}
		close ix;
	    }
	    elsif($tmp[0] eq $F_GENEBANK){
		$HANDLE{GenebankParser->type} = GenebankParser->parse($tmp[1]);
		$HANDLE{GenebankParser->type}->sort_elements;
	    }
	    elsif ($tmp[0] eq $F_ZCURVE){
		$HANDLE{ZCurveParser->type} = ZCurveParser->parse($tmp[1]);
		$HANDLE{ZCurveParser->type}->sort_elements;
	    }
	    elsif ($tmp[0] eq $F_CRITICA){
		$HANDLE{CriticaParser->type} = CriticaParser->parse($tmp[1]);
		$HANDLE{CriticaParser->type}->sort_elements;
	    }
	    elsif($tmp[0] eq $F_GLIMMER_RBS){
		$HANDLE{GlimmerRBSParser->type} = GlimmerRBSParser->parse(
                                                                 $tmp[1]);
		$HANDLE{GlimmerRBSParser->type}->sort_elements;
	    }
	    elsif($tmp[0] eq $F_MINLENGTH){
		if ( $tmp[1]){
		    $MINLENGTH=$tmp[1];
		}
	    }
	    else{print $USAGE;exit(1);}
	}
	else {print $USAGE;exit(1);}
    }
}
